@@ -11,21 +11,22 @@
 # License for the specific language governing permissions and limitations
 # under the License.
 
-"""Tests for ceilometer/polling/dynamic_pollster.py"""
+"""Tests for OpenStack dynamic pollster
+"""
 import copy
 import logging
 from unittest import mock
 
-from oslotest import base
 import requests
+from six.moves.urllib import parse as url_parse
 
 from ceilometer.declarative import DynamicPollsterDefinitionException
 from ceilometer.polling import dynamic_pollster
 from ceilometer import sample
+from oslotest import base
 
 LOG = logging.getLogger(__name__)
 
-
 REQUIRED_POLLSTER_FIELDS = ['name', 'sample_type', 'unit',
                             'value_attribute', 'endpoint_type',
                             'url_path']
@@ -75,7 +76,7 @@ class PagedSamplesGenerator(SampleGenerator):
         for page_link, page_size in page_links.items():
             page_link = page_base_link + "/" + page_link
             self.response[current_page_link] = {
-                self.page_link_name: page_link,
+                self.page_link_name: [{'href': page_link, 'rel': 'next'}],
                 self.dict_name: self.populate_page(page_size)
             }
             current_page_link = page_link
@@ -155,7 +156,7 @@ class TestDynamicPollster(base.BaseTestCase):
         self.assertEqual(pollster_definition, pollster.pollster_definitions)
 
     @mock.patch('keystoneclient.v2_0.client.Client')
-    def test_skip_samples(self, keystone_mock):
+    def test_skip_samples_with_linked_samples(self, keystone_mock):
         generator = PagedSamplesGeneratorHttpRequestMock(samples_dict={
             'volume': SampleGenerator(samples_dict={
                 'name': ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h'],
@@ -179,16 +180,26 @@ class TestDynamicPollster(base.BaseTestCase):
         pollster_definition['skip_sample_values'] = ['rb']
         pollster_definition['url_path'] = 'v1/test-volumes'
         pollster_definition['response_entries_key'] = 'servers'
+        pollster_definition['next_sample_url_attribute'] = \
+            'server_link | filter(lambda v: v.get("rel") == "next", value) |' \
+            'list(value) | value[0] | value.get("href")'
         pollster = dynamic_pollster.DynamicPollster(pollster_definition)
         samples = pollster.get_samples(fake_manager, None, ['http://test.com'])
-        self.assertEqual(['ra', 'rc'], list(map(lambda s: s.volume, samples)))
+        self.assertEqual(['ra', 'rc', 'rd', 're', 'rf', 'rg', 'rh'],
+                         list(map(lambda s: s.volume, samples)))
+
+        generator.generate_samples('http://test.com/v1/test-volumes', {
+            'marker=c3': 3,
+            'marker=f6': 3
+        }, 2)
 
         pollster_definition['name'] = 'test-pollster'
         pollster_definition['value_attribute'] = 'name'
         pollster_definition['skip_sample_values'] = ['b2']
         pollster = dynamic_pollster.DynamicPollster(pollster_definition)
         samples = pollster.get_samples(fake_manager, None, ['http://test.com'])
-        self.assertEqual(['a1', 'c3'], list(map(lambda s: s.volume, samples)))
+        self.assertEqual(['a1', 'c3', 'd4', 'e5', 'f6', 'g7', 'h8'],
+                         list(map(lambda s: s.volume, samples)))
 
     def test_all_required_fields_ok(self):
         pollster = dynamic_pollster.DynamicPollster(
@@ -704,3 +715,285 @@ class TestDynamicPollster(base.BaseTestCase):
         self.assertEqual(46, get_obj_sample.volume)
         self.assertEqual(8, list_bucket_sample.volume)
         self.assertEqual(46, put_obj_sample.volume)
+
+    def test_execute_request_get_samples_custom_ids(self):
+        sample = {'user_id_attribute': "1",
+                  'project_id_attribute': "2",
+                  'resource_id_attribute': "3",
+                  'user_id': "234",
+                  'project_id': "2334",
+                  'id': "35"}
+
+        def internal_execute_request_get_samples_mock(self, arg):
+            class Response:
+                def json(self):
+                    return [sample]
+            return Response(), "url"
+
+        original_method = dynamic_pollster.PollsterSampleGatherer.\
+            internal_execute_request_get_samples
+        try:
+            dynamic_pollster.PollsterSampleGatherer. \
+                internal_execute_request_get_samples = \
+                internal_execute_request_get_samples_mock
+
+            self.pollster_definition_all_fields[
+                'user_id_attribute'] = 'user_id_attribute'
+            self.pollster_definition_all_fields[
+                'project_id_attribute'] = 'project_id_attribute'
+            self.pollster_definition_all_fields[
+                'resource_id_attribute'] = 'resource_id_attribute'
+
+            pollster = dynamic_pollster.DynamicPollster(
+                self.pollster_definition_all_fields)
+
+            params = {"d": "d"}
+            response = pollster.definitions.sample_gatherer. \
+                execute_request_get_samples(**params)
+
+            self.assertEqual(sample['user_id_attribute'],
+                             response[0]['user_id'])
+            self.assertEqual(sample['project_id_attribute'],
+                             response[0]['project_id'])
+            self.assertEqual(sample['resource_id_attribute'],
+                             response[0]['id'])
+        finally:
+            dynamic_pollster.PollsterSampleGatherer. \
+                internal_execute_request_get_samples = original_method
+
+    def test_retrieve_attribute_self_reference_sample(self):
+        key = " . | value['key1']['subKey1'][0]['d'] if 'key1' in value else 0"
+
+        sub_value1 = [{"d": 2}, {"g": {"h": "val"}}]
+        sub_value2 = [{"r": 245}, {"h": {"yu": "yu"}}]
+
+        json_object = {"key1": {"subKey1": sub_value1},
+                       "key2": {"subkey2": sub_value2}}
+
+        pollster = dynamic_pollster.DynamicPollster(
+            self.pollster_definition_only_required_fields)
+
+        returned_value = pollster.definitions.sample_extractor.\
+            retrieve_attribute_nested_value(json_object, key)
+        self.assertEqual(2, returned_value)
+
+        del json_object['key1']
+        returned_value = pollster.definitions.sample_extractor.\
+            retrieve_attribute_nested_value(json_object, key)
+        self.assertEqual(0, returned_value)
+
+    def test_create_request_arguments_NonOpenStackApisSamplesGatherer(self):
+        pollster_definition = {
+            'name': "test-pollster", 'sample_type': "gauge", 'unit': "test",
+            'value_attribute': "volume",
+            'url_path': "https://test.com/v1/test/endpoint/fake",
+            "module": "someModule",
+            "authentication_object": "objectAuthentication",
+            "authentication_parameters": "authParam",
+            "headers": [{"header1": "val1"}, {"header2": "val2"}]}
+
+        pollster = dynamic_pollster.DynamicPollster(pollster_definition)
+
+        request_args = pollster.definitions.sample_gatherer\
+            .create_request_arguments()
+
+        self.assertTrue("headers" in request_args)
+        self.assertEqual(2, len(request_args["headers"]))
+
+        self.assertEqual(['header1', 'header2'],
+                         list(map(lambda h: list(h.keys())[0],
+                                  request_args["headers"])))
+
+        self.assertEqual(['val1', 'val2'],
+                         list(map(lambda h: list(h.values())[0],
+                                  request_args["headers"])))
+
+        self.assertTrue("authenticated" not in request_args)
+
+    def test_create_request_arguments_PollsterSampleGatherer(self):
+        pollster_definition = copy.deepcopy(
+            self.pollster_definition_only_required_fields)
+        pollster_definition["headers"] = [
+            {"x-openstack-nova-api-version": "2.46"},
+            {"custom_header": "custom"},
+            {"some_other_header": "something"}]
+
+        pollster = dynamic_pollster.DynamicPollster(pollster_definition)
+
+        request_args = pollster.definitions.sample_gatherer\
+            .create_request_arguments()
+
+        self.assertTrue("headers" in request_args)
+        self.assertTrue("authenticated" in request_args)
+        self.assertTrue(request_args["authenticated"])
+
+        self.assertEqual(3, len(request_args["headers"]))
+
+        self.assertEqual(['x-openstack-nova-api-version', 'custom_header',
+                          "some_other_header"],
+                         list(map(lambda h: list(h.keys())[0],
+                                  request_args["headers"])))
+
+        self.assertEqual(['2.46', 'custom', 'something'],
+                         list(map(lambda h: list(h.values())[0],
+                                  request_args["headers"])))
+
+    def test_create_request_arguments_PollsterSampleGatherer_no_headers(self):
+        pollster = dynamic_pollster.DynamicPollster(
+            self.pollster_definition_only_required_fields)
+
+        request_args =\
+            pollster.definitions.sample_gatherer.create_request_arguments()
+
+        self.assertTrue("headers" not in request_args)
+        self.assertTrue("authenticated" in request_args)
+        self.assertTrue(request_args["authenticated"])
+
+    @mock.patch('keystoneclient.v2_0.client.Client')
+    def test_metadata_nested_objects(self, keystone_mock):
+        generator = PagedSamplesGeneratorHttpRequestMock(samples_dict={
+            'flavor': [{"name": "a", "ram": 1}, {"name": "b", "ram": 2},
+                       {"name": "c", "ram": 3}, {"name": "d", "ram": 4},
+                       {"name": "e", "ram": 5}, {"name": "f", "ram": 6},
+                       {"name": "g", "ram": 7}, {"name": "h", "ram": 8}],
+            'name': ['s1', 's2', 's3', 's4', 's5', 's6', 's7', 's8'],
+            'state': ['Active', 'Error', 'Down', 'Active', 'Active',
+                      'Migrating', 'Active', 'Error']
+        }, dict_name='servers', page_link_name='server_link')
+
+        generator.generate_samples('http://test.com/v1/test-servers', {
+            'marker=c3': 3,
+            'marker=f6': 3
+        }, 2)
+
+        keystone_mock.session.get.side_effect = generator.mock_request
+        fake_manager = self.FakeManager(keystone=keystone_mock)
+
+        pollster_definition = dict(self.multi_metric_pollster_definition)
+        pollster_definition['name'] = 'test-pollster'
+        pollster_definition['value_attribute'] = 'state'
+        pollster_definition['url_path'] = 'v1/test-servers'
+        pollster_definition['response_entries_key'] = 'servers'
+        pollster_definition['metadata_fields'] = ['flavor.name', 'flavor.ram']
+        pollster_definition['next_sample_url_attribute'] = \
+            'server_link | filter(lambda v: v.get("rel") == "next", value) |' \
+            'list(value)| value [0] | value.get("href")'
+        pollster = dynamic_pollster.DynamicPollster(pollster_definition)
+
+        samples = pollster.get_samples(fake_manager, None, ['http://test.com'])
+
+        samples = list(samples)
+
+        self.assertEqual(['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h'],
+                         list(map(lambda s: s.resource_metadata["flavor.name"],
+                                  samples)))
+        self.assertEqual(list(range(1, 9)),
+                         list(map(lambda s: s.resource_metadata["flavor.ram"],
+                                  samples)))
+
+    def test_get_request_linked_samples_url_no_next_sample(self):
+        pollster = dynamic_pollster.DynamicPollster(
+            self.pollster_definition_only_required_fields)
+
+        base_url = "http://test.com/something_that_we_do_not_care"
+        expected_url = url_parse.urljoin(
+            base_url, self.pollster_definition_only_required_fields[
+                'url_path'])
+
+        kwargs = {'resource': base_url}
+        url = pollster.definitions.sample_gatherer\
+            .get_request_linked_samples_url(kwargs)
+
+        self.assertEqual(expected_url, url)
+
+    def test_get_request_linked_samples_url_next_sample_url(self):
+        pollster = dynamic_pollster.DynamicPollster(
+            self.pollster_definition_only_required_fields)
+
+        base_url = "http://test.com/something_that_we_do_not_care"
+        expected_url = "http://test.com/next_page"
+
+        kwargs = {'resource': base_url,
+                  'next_sample_url': expected_url}
+
+        url = pollster.definitions.sample_gatherer\
+            .get_request_linked_samples_url(kwargs)
+
+        self.assertEqual(expected_url, url)
+
+    def test_get_request_linked_samples_url_next_sample_only_url_path(self):
+        pollster = dynamic_pollster.DynamicPollster(
+            self.pollster_definition_only_required_fields)
+
+        base_url = "http://test.com/something_that_we_do_not_care"
+        expected_url = "http://test.com/next_page"
+
+        kwargs = {'resource': base_url,
+                  'next_sample_url': "/next_page"}
+
+        url = pollster.definitions.sample_gatherer\
+            .get_request_linked_samples_url(kwargs)
+
+        self.assertEqual(expected_url, url)
+
+    def test_generate_sample_and_extract_metadata(self):
+        definition = self.pollster_definition_only_required_fields.copy()
+        definition['metadata_fields'] = ["metadata1", 'metadata2']
+
+        pollster = dynamic_pollster.DynamicPollster(definition)
+
+        pollster_sample = {'metadata1': 'metadata1',
+                           'metadata2': 'metadata2',
+                           'value': 1}
+
+        sample = pollster.definitions.sample_extractor.generate_sample(
+            pollster_sample)
+
+        self.assertEqual(1, sample.volume)
+        self.assertEqual(2, len(sample.resource_metadata))
+        self.assertEqual('metadata1', sample.resource_metadata['metadata1'])
+        self.assertEqual('metadata2', sample.resource_metadata['metadata2'])
+
+    def test_generate_sample_and_extract_metadata_false_value(self):
+        definition = self.pollster_definition_only_required_fields.copy()
+        definition['metadata_fields'] = ["metadata1",
+                                         'metadata2',
+                                         'metadata3_false']
+
+        pollster = dynamic_pollster.DynamicPollster(definition)
+
+        pollster_sample = {'metadata1': 'metadata1',
+                           'metadata2': 'metadata2',
+                           'metadata3_false': False,
+                           'value': 1}
+
+        sample = pollster.definitions.sample_extractor.generate_sample(
+            pollster_sample)
+
+        self.assertEqual(1, sample.volume)
+        self.assertEqual(3, len(sample.resource_metadata))
+        self.assertEqual('metadata1', sample.resource_metadata['metadata1'])
+        self.assertEqual('metadata2', sample.resource_metadata['metadata2'])
+        self.assertIs(False, sample.resource_metadata['metadata3_false'])
+
+    def test_generate_sample_and_extract_metadata_none_value(self):
+        definition = self.pollster_definition_only_required_fields.copy()
+        definition['metadata_fields'] = ["metadata1",
+                                         'metadata2',
+                                         'metadata3']
+
+        pollster = dynamic_pollster.DynamicPollster(definition)
+
+        pollster_sample = {'metadata1': 'metadata1',
+                           'metadata2': 'metadata2',
+                           'metadata3': None,
+                           'value': 1}
+
+        sample = pollster.definitions.sample_extractor.generate_sample(
+            pollster_sample)
+
+        self.assertEqual(1, sample.volume)
+        self.assertEqual(3, len(sample.resource_metadata))
+        self.assertEqual('metadata1', sample.resource_metadata['metadata1'])
+        self.assertEqual('metadata2', sample.resource_metadata['metadata2'])
+        self.assertIsNone(sample.resource_metadata['metadata3'])
